#include "version.h"
#include "winres.h"

RCFILE_FC_JSON_ELASTIC_IMPORT RCDATA DISCARDABLE "res\\m_fc_json_elastic_import.ps1"
RCFILE_FC_JSON_ELASTIC_IMPORT_UNAUTH RCDATA DISCARDABLE "res\\m_fc_json_elastic_import_unauth.ps1"

/////////////////////////////////////////////////////////////////////////////
// Neutral (Default) (unknown sub-lang: 0x8) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ZZZ)
LANGUAGE LANG_NEUTRAL, 0x8

/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION VER_FILE_VERSION
 PRODUCTVERSION VER_PRODUCT_VERSION
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040004b0"
        BEGIN
            VALUE "CompanyName", VER_COMPANY_NAME_STR "\0"
            VALUE "FileDescription", VER_FILE_DESCRIPTION_STR "\0"
            VALUE "FileVersion", VER_FILE_VERSION_STR "\0"
            VALUE "InternalName", VER_INTERNAL_NAME_STR "\0"
            VALUE "LegalCopyright", VER_COPYRIGHT_STR "\0"
            VALUE "OriginalFilename", VER_ORIGINAL_FILENAME_STR "\0"
            VALUE "ProductName", VER_PRODUCTNAME_STR "\0"
            VALUE "ProductVersion", VER_PRODUCT_VERSION_STR "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x400, 1200
    END
END

#endif    // Neutral (Default) (unknown sub-lang: 0x8) resources
/////////////////////////////////////////////////////////////////////////////

